match x:
    case var: ...  # capture pattern
    case 2: ...
match x:
    case _: ...
    case 2: ...    # wildcard pattern
match x:
    case var1 as var2: ...  # as pattern with irrefutable left-hand side
    case 2: ...
match x:
    case enum.variant | var: ...  # or pattern with irrefutable part
    case 2: ...
