# parse_options: {"target-version": "3.15"}
try:
    lazy import os
except:
    pass

try:
    x
except* Exception:
    lazy import sys

def func():
    lazy import math

async def async_func():
    lazy from json import loads

class MyClass:
    lazy import typing

def outer():
    class Inner:
        lazy import json
