# parse_options: {"target-version": "3.10"}
# regression tests for https://github.com/astral-sh/ruff/issues/16874
# starred parameters are fine, just not the annotation
from typing import Annotated, Literal
def foo(*args: Ts): ...
def foo(*x: Literal["this should allow arbitrary strings"]): ...
def foo(*x: Annotated[str, "this should allow arbitrary strings"]): ...
def foo(*args: str, **kwds: int): ...
def union(*x: A | B): ...
