match ruff:
    case {"lint": {"select": x} | {"extend-select": x}} | {"select": x}:
        ...
match 42:
    case [[x] | [x]] | x: ...
match 42:
    case [[x | x] | [x]] | x: ...
match 42:
    case ast.Subscript(n, ast.Constant() | ast.Slice()) | ast.Attribute(n): ...
